/*********************************************************************
Copyright(C) 1994,1995 by Vladimir Belkin
All Rights Reserved

E-mail: vladimir.belkin@usa.net

You are free to use this file for any purposes.
*********************************************************************/

#ifndef __EXCEPTION_ARTIFIC_HANDLING__
#define __EXCEPTION_ARTIFIC_HANDLING__

#include <new.h>
#include <setjmp.h>

#define DeclareTypeId(type,xx) \
  inline int ptypeId(const type*) {return (xx);} \
  inline int typeId(const type&) {return (xx);}

#define typeTypeId(T) ptypeId((T *)NULL)

#define NULL_ASN(T) \
T &operator = (const T &) {return * this;}

#define xArt

class Checked;
extern Checked  *TheLastChecked;
class TryToExecute;
extern TryToExecute  *TheLastTryToExecute;
extern int TheLastExceptionType;
extern void *TheLastExceptionLvalue;

#define MAX_EXCEPTION_SIZE 0x40

extern char exceptionBuffer[MAX_EXCEPTION_SIZE];

#ifndef _WINDOWS
  extern unsigned TheStackBottom;
  void SetTheStackBottom(void  *p);
#endif

#ifdef _MSC_VER	

inline void *operator new(unsigned, void *p) {
  return p;
}

typedef void (* PFVV)();

#define _JBLEN  9

typedef  int  jmp_buf[_JBLEN];

extern "C" {
	int  __cdecl _setjmp(jmp_buf);
	void __cdecl longjmp(jmp_buf, int);
}

#define setjmp  _setjmp

#endif


extern "C" {
 void  terminate();
 PFVV  set_terminate(PFVV);
}



class Checked {
  friend class TryToExecute;
  friend void OnException(int type,void *p);
  Checked  *prev;
public:

  NULL_ASN(Checked)
  Checked(const Checked&) { new(this) Checked(); }

  void DestroyChecked();
  virtual ~Checked();
  Checked();
  void resetChecked();
};



class TryToExecute {
public:

  jmp_buf buf;
  Checked  *chk;
  TryToExecute  *tr;

  int type;
  void *lvalue;

  void OnError(int type,void *ptr);

  TryToExecute();
  ~TryToExecute();

};

void checkExceptionSize(int,int);
void OnException(int type,void *p);

#define DeclareException(T,Id) \
DeclareTypeId(T,Id) \
inline void xThrow(T value) \
{ \
  checkExceptionSize(sizeof(T),Id); \
  new(exceptionBuffer) T(value); \
  OnException(Id,exceptionBuffer); \
}

DeclareException(int,1)
DeclareException(int*,10)
DeclareException(int**,11)
DeclareException(int***,12)
DeclareException(unsigned int,2)
DeclareException(unsigned int*,20)
DeclareException(unsigned int**,21)
DeclareException(unsigned int***,22)
DeclareException(short,1)
DeclareException(unsigned short,2)
DeclareException(long,3)
DeclareException(unsigned long,4)
DeclareException(float,5)
DeclareException(double,6)
DeclareException(char,7)
DeclareException(unsigned char,8)
DeclareException(short*,10)
DeclareException(unsigned short*,20)
DeclareException(long*,30)
DeclareException(unsigned long*,40)
DeclareException(float*,50)
DeclareException(double*,60)
DeclareException(char*,70)
DeclareException(unsigned char*,80)
DeclareException(short**,11)
DeclareException(unsigned short**,21)
DeclareException(long**,31)
DeclareException(unsigned long**,41)
DeclareException(float**,51)
DeclareException(double**,61)
DeclareException(char**,71)
DeclareException(unsigned char**,81)
DeclareException(short***,12)
DeclareException(unsigned short***,22)
DeclareException(long***,32)
DeclareException(unsigned long***,42)
DeclareException(float***,52)
DeclareException(double***,62)
DeclareException(char***,72)
DeclareException(unsigned char***,82)


struct XX_invalid_size {
  int type;
  XX_invalid_size(int tid):type(tid) {}
};

DeclareException(XX_invalid_size,30000)

void xThrow(void);

#define xTry {TryToExecute _TTE_;if (!setjmp(_TTE_.buf)) {
#define xCatch(t,val) } else if (typeTypeId(t)==_TTE_.type) {\
             t &val=*(t *)(_TTE_.lvalue);_TTE_.type=0;
#define xCatchAll } else {_TTE_.type=0;
#define xCatchType(t) } else if (typeTypeId(t)==_TTE_.type) {\
   _TTE_.type=0;
#define xTryEnd }}

#define COLON_CHECKED :protected Checked
#define CHECKED_COMMA protected Checked,
#define COLON_VCHECKED :protected virtual Checked
#define VCHECKED_COMMA protected virtual Checked,

#endif
